/* $Header: /cvs/Darwin/src/live/tcsh/tcsh/snames.h,v 1.1.1.2 2001/06/28 23:10:53 bbraun Exp $ */
/*
 * snames.h: Short names for old compilers
 */
#ifndef _snames_h
#define _snames_h

#define	tw_cmd_add		twcmdadd
#define	tw_cmd_alias		twcmdalias
#define	tw_cmd_sort		twcmdsort
#define	tw_cmd_start		twcmdstart
#define	tw_logname_end		twnameend
#define	tw_logname_next		twnamenext
#define	tw_logname_start	twnamestart
#define	tw_file_next		twfilenext
#define	tw_file_start		twfilestart
#define	tw_item_add		twitemadd
#define	tw_item_find		twitemfind
#define	tw_item_get		twitemget
#define	tw_bind_next		twbindnext
#define	tw_bind_start		twbindstart
#define	tw_limit_next		twlimitnext
#define	tw_limit_start		twlimitstart
#define	tw_collect		twcllct
#define	tw_collect_items	twcllcti
#define	e_complete		ecomp
#define	e_complete_all		ecompall
#define	e_complete_back		ecompback
#define	e_complete_fwd		ecompfwd
#define	e_correct		ecrrct
#define	e_correctl		ecrrctl
#define	e_expand_glob		eexpandglob
#define	e_expand_vars		eexpandvars
#define	e_delnext		edelnxt
#define	e_delnext_eof		edelnxteof
#define e_delnext_list		edelnxtlst
#define e_delnext_list_eof	edellsteof
#define	e_delwordnext		edelwrdnext
#define	e_delwordprev		edelwrdprev
#define	e_expand		eexp
#define	e_expand_history	eexphistory
#define	e_expand_vars		eexpvars
#define	e_tty_starto		ettystarto
#define	e_tty_stopo		ettystopo
#define	v_csearch_back		vcsrchback
#define	v_csearch_fwd		vcsrchfwd
#define	v_charto_back		vchartovack
#define	v_charto_fwd		vchartofwd
#define	v_rchar_back		vrcharback
#define	v_rchar_fwd		vrcharfwd
#define	v_rsrch_back		vrsrchback
#define	v_rsrch_fwd		vrsrchfwd
#define	ed_InitMaps		edInitMaps
#define	ed_InitMetaBindings	edInitMetaBindings
#define	QuoteModeOff		QuoteOff
#define	QuoteModeOn		QuoteOn
#define	GetNextChar		GetNxtChar
#define	GetNextCommand		GetNxtCommand
#define	tty_gettabs		ttygettabs
#define	tty_getty		ttygetty
#define	printkey		prntkey
#define	printkeys		prntkeys
#define	STRprompt		STRpmpt
#define	STRprompt2		STRpmpt2
#define	STRprompt3		STRpmpt3
#define	STRtildotdirs		STRtdotdirs
#define	STRtildothist		STRtdothist
#define	STRpushdsilent		STRpushsilent
#define	STRpushdtohome		STRpushtohome
#define	STRdefault		STRdef
#define	STRdefautologout	STRdefautologout
#define	STRfakecom		STRfake
#define	STRfakecom1		STRfake1
#define	STRLparen		STRLpar
#define	STRLparensp		STRLparsp
#define	STRspLarrowsp		STRspLarrowsp
#define	STRspLarrow2sp		STRspL2arrowsp
#define	STRspRarrow		STRspRarrow
#define	STRspRarrow2		STRspR2arrow
#define	STRcontinue		STRcont
#define	STRcontinue_args	STRcont_args
#define	STRsldotcshrc		STRsdotcshrc
#define	STRsldottcshrc		STRsdottcshrc
#define	STRsldotlogin		STRsdotlogin
#define	STRlistmax		STRlmax
#define	STRlistmaxrows		STRlmaxrows

#endif /* _snames_h */
